__author__ = 'ShortIO Team'
__author_email__ = 'shortio@protonmail.com'
__description__ = 'ShortIO library is intended to avoid context manager boilerplate in simple io operations.'
__license__ = 'MIT'
__title__ = 'shortio'
__url__ = 'https://github.com/byshyk/shortio'
__version__ = '0.1.0'
